! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2015 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                          Test Program                        !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

program test

  use timing_mod
  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_block_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_util_mod
  use fd4_balance_mod
  use fd4_boundary_mod
#ifdef FD4_VIS5D
  use fd4_vis5d_mod
#endif

  USE_MPI
  
  implicit none
  
  INCLUDE_MPIFH


  integer :: rank, nproc, err
  
  call MPI_Init(err)
  call MPI_Comm_size(MPI_COMM_WORLD,nproc,err)
  call MPI_Comm_rank(MPI_COMM_WORLD,rank ,err) 
  
  !call test_put_array(4, (/1,1,1/), (/.true.,.true.,.true./), 0, MPI_COMM_WORLD, .true., 1 )

  !call test_put_array(4, (/1,1,1/), (/.true.,.true.,.true./), 1, MPI_COMM_WORLD, .true., 1 )
  !call test_put_array(4, (/2,2,2/), (/.true.,.true.,.true./), 1, MPI_COMM_WORLD, .true., 1 )
  !call test_put_array(4, (/0,0,0/), (/.true.,.true.,.true./), 1, MPI_COMM_WORLD, .true., 1 )
  !call test_put_array(32, (/2,2,2/), (/.true.,.true.,.true./), 1, MPI_COMM_WORLD, .false., 1 )
  !call test_put_array(32, (/2,2,4/), (/.true.,.true.,.true./), 1, MPI_COMM_WORLD, .false., 50 )
  
  !call test_put_array(8, (/1,1,1/), (/.true.,.true.,.true./), 2, MPI_COMM_WORLD, .true., 1 )
  !call test_put_array(8, (/0,0,0/), (/.true.,.true.,.true./), 2, MPI_COMM_WORLD, .true., 1 )
  !call test_put_array(32, (/1,1,2/), (/.true.,.true.,.true./), 2, MPI_COMM_WORLD, .false., 50 )
  
  !call test_put_array(32, (/1,1,2/), (/.true.,.true.,.true./), 3, MPI_COMM_WORLD, .false., 500 )
  !call test_put_array(32, (/1,1,2/), (/.true.,.true.,.true./), 3, MPI_COMM_WORLD, .false., 10 )
  !call test_put_array(8, (/1,1,2/), (/.true.,.true.,.true./), 3, MPI_COMM_WORLD, .false., 10 )
  !call test_put_array(64, (/4,4,8/), (/.true.,.true.,.true./), 3, MPI_COMM_WORLD, .false., 1 )
  
  !call test_put_array(16, (/1,1,1/), (/.true.,.true.,.true./), 4, MPI_COMM_WORLD, .true., 50 )

  !call test_put_array(4, (/1,1,1/), (/.true.,.true.,.true./), 5, MPI_COMM_WORLD, .true., 1 )
  !call test_put_array(8, (/1,1,1/), (/.true.,.true.,.true./), 5, MPI_COMM_WORLD, .true., 20 )
  
  !call test_put_array(4, (/1,1,1/), (/.true.,.true.,.true./), 6, MPI_COMM_WORLD, .true., 20 )
  !call test_put_array(16, (/4,4,4/), (/.true.,.true.,.true./), 6, MPI_COMM_WORLD, .true., 1 )


  !call test_put_array(4, (/1,1,0/), (/.true.,.true.,.true./), 7, MPI_COMM_WORLD, .true., 10 )
  !call test_put_array(6, (/2,2,2/), (/.true.,.true.,.true./), 7, MPI_COMM_WORLD, .true., 1 )

  call test_put_array(4, (/1,1,1/), (/.true.,.true.,.true./), 8, MPI_COMM_WORLD, .true., 10 )
  !call test_put_array(4, (/4,4,4/), (/.true.,.true.,.true./), 8, MPI_COMM_WORLD, .true., 1 )

  
  !call test_put_array(4, (/1,1,1/), (/.true.,.true.,.true./), 9, MPI_COMM_WORLD, .true., 10 )
  !call test_put_array(6, (/2,2,2/), (/.true.,.true.,.true./), 9, MPI_COMM_WORLD, .true., 1 )


  call MPI_Finalize(err)

end program test




! some tests for parallel fd4_util_put_array
! minimun sz = 4
! scenarios:
!  - 0: one thin "line cloud"
!  - 1: two thin "line clouds"
!  - 2: two spheric clouds
!  - 3: a moving cloud
!  - 4: single block case of scenario 2
!  - 5: remove data (putting zeros - deleting blocks)
!  - 6: two single grid cells
!  - 7: test for iter_get_ghost
!  - 8: test for fd4_couple_put
!  - 9: test for fd4_couple_mark_blocks - fd4_balance_readjust - fd4_couple_put
!
subroutine test_put_array(sz, ng, periodic, scenario, comm, dump, bins)
  use fd4_mod
  use fd4_couple_mod, only: fd4_couple_debug_metadata
  implicit none

  integer(i_k), intent(in) :: sz
  integer(i_k), intent(in) :: ng(3)
  logical, intent(in) :: periodic(3)
  integer(i_k), intent(in) :: scenario
  integer(i_k), intent(in) :: comm
  logical, intent(in) :: dump
  integer(i_k), intent(in) :: bins


  type(fd4_vartab) :: vartab(1:3)
  integer(i_k) :: err
  integer(i_k) :: dsize(3,2), bsize(3), abnd(0:3,2), aext(0:3), abnd2(0:3,2), o(3)
  type(fd4_domain), target :: d
  type(fd4_iter) :: iter
  real(r_k), allocatable :: array(:,:,:,:), compare(:,:,:,:), array2(:,:,:,:)
  integer :: rank, nproc
  character(len=128) :: str,str2
  integer(i_k) :: steps, s, i
  real :: move

  real(r_k), allocatable :: work(:,:,:,:)
  integer(i_k) :: wext(0:3), bext(0:3), k
  type(fd4_ghostcomm) :: gc
  
  type(fd4_couple) :: cpl
  integer(i_k) :: cabnd(3,2), caidx(2)
  real(r_k), allocatable, target :: cplarray(:,:,:), cplarray4(:,:,:,:), cplarrayf(:,:,:)
  real(r_k), pointer :: cplarrayp(:,:,:), cplarray4p(:,:,:,:)


  call MPI_Comm_size(comm,nproc,err)
  call MPI_Comm_rank(comm,rank ,err) 

  dsize(:,1) = (/ 1, 1, 1/)
  dsize(:,2) = (/ sz, sz, sz/)
  o(:) = 0
  dsize(:,1) = dsize(:,1) + o(:)
  dsize(:,2) = dsize(:,2) + o(:)
  bsize(:) = dsize(:,2) - dsize(:,1) + 1
  where(ng>1) bsize=bsize/ng
  if(scenario==4) bsize(:) = 1

  vartab(1) = fd4_vartab('Test', bins, 2, .false., 0.0_8, 1.e-20_8      , 0 )
  vartab(2) = fd4_vartab('U'   , 1   , 1, .false., 2.0_8, FD4_NOTHRES   , 1 )
  vartab(3) = fd4_vartab('Foo' , 1   , 1, .false., 0.0_8, FD4_NOTHRES   , 0 )

  call fd4_domain_create(d, bsize, dsize, vartab, ng, periodic, comm, err)
  if(err/=0) stop

  steps=2
  if(scenario==3) steps=16  ! 96
  if(scenario==5) steps=3
  if(scenario==0) steps=1
  if(scenario==7) steps=1
  if(scenario==8) steps=1
  if(scenario==9) steps=1

#ifdef FD4_VIS5D
  call fd4_vis5d_open(d,'out.v5d',steps,2,(/1,2/),(/1,1/),err)  
#endif  

  abnd(1:3,:) = dsize(1:3,:)
  abnd(0,:) = (/ 1, bins /)
  aext(:) = abnd(:,2)-abnd(:,1)+1
  allocate(array(aext(0),aext(1),aext(2),aext(3)))
  array = 0.0_r_k
  allocate(compare(aext(0),aext(1),aext(2),aext(3)))
  compare = 0.0_r_k

  if(scenario==0) then

    array(:,sz/4:sz/4,sz/4:sz/2,sz/2:sz/2) = 1.0_r_k

    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump

!#ifdef BLA
    call fd4_domain_max_bext(d,wext(1:3),.true.)
    wext(0)=bins
    allocate(work(wext(0),wext(1),wext(2),wext(3)))
    work=0.0_r_k
    call fd4_iter_init(d, iter)
    do while(associated(iter%cur))
      !if(iter%cur%pos(1)==1 .and. iter%cur%pos(2)==1 .and. iter%cur%pos(3)==2) then
      if(iter%cur%pos(1)==1 .and. iter%cur%pos(2)==1 .and. iter%cur%pos(3)==1) then
        work = 1.0_r_k
        call fd4_iter_put_ghost(iter,1,1,wext,work,.true.)
      end if
      call fd4_iter_next(iter)
    end do
!#endif

    call fd4_balance_readjust(d,err)
    !call fd4_balance_readjust(d,err,FD4_BALANCE_METIS)
    !call get_compare_dump
    
  elseif(scenario==1) then

    array(:,sz/4:sz/4,sz/4:sz/2,sz/2:sz/2) = 1.0_r_k

    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump

    array = 0.0_r_k
    array(:,3*sz/4:3*sz/4,3*sz/4:sz,sz:sz) = 1.0_r_k

    abnd(1,:) = (/ 3*sz/4 , 3*sz/4 /)
    call fd4_util_put_array(d, 1, 1, abnd, array(:,3*sz/4:3*sz/4,:,:), err)
    abnd(1:3,:) = dsize(1:3,:)
    print *,rank,' number of blocks: ',d%blocks%length


    array(:,sz/4:sz/4,sz/4:sz/2,sz/2:sz/2) = 1.0_r_k
    call get_compare_dump

    !call fd4_balance_readjust(d,err,FD4_BALANCE_METIS)


  elseif(scenario==2 .or. scenario==4) then


    call mkcloud(array,1.0_r_k,REAL(sz)/4.0,REAL(sz)/4.0,REAL(sz)/4.0,MAX(1.0,REAL(sz)/4.0))

    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump

    array = 0.0_r_k
    call mkcloud(array,1.0_r_k,3.0*REAL(sz)/4.0,3.0*REAL(sz)/4.0,3.0*REAL(sz)/4.0,MAX(1.0,REAL(sz)/4.0))

    abnd(1,:) = (/ sz/2+1 , sz /)
    abnd(2,:) = (/ sz/2+1 , sz /)
    call fd4_util_put_array(d, 1, 1, abnd, array(:,sz/2+1:sz,sz/2+1:sz,:), err)
    abnd(1:3,:) = dsize(1:3,:)
    print *,rank,' number of blocks: ',d%blocks%length


    call mkcloud(array,1.0_r_k,REAL(sz)/4.0,REAL(sz)/4.0,REAL(sz)/4.0,MAX(1.0,REAL(sz)/4.0))
    call get_compare_dump
    
  elseif(scenario==3) then
  
    move = REAL(sz)/REAL(2*steps)
  
    do s=1,steps
      array = 0.0_r_k
      call mkcloud(array,1.0_r_k,REAL(sz)/4.0+move*REAL(s),REAL(sz)/4.0+move*REAL(s),REAL(sz)/2.0,MAX(1.0,REAL(sz)/8.0))
      !array(:,s:s+4,s+2:s+7,6:15) = 1.0_r_k
      call fd4_util_put_array(d, 1, 1, abnd, array, err)
      !print *,rank,' number of blocks: ',d%blocks%length
      call get_compare_dump
    end do

  elseif(scenario==5) then

    array = 1.0_r_k
    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump

    array = 1.0_r_k
    call mkcloud(array,0.0_r_k,REAL(sz)/2.0,REAL(sz)/2.0,REAL(sz)/2.0,REAL(sz)/2.0)
    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump

    array = 0.0_r_k
    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump

  elseif(scenario==6) then
    
    allocate(array2(1:bins,1,1,1))
    abnd2(0:3,1) = (/ 1,    2, 2, 2/)
    abnd2(0:3,2) = (/ bins, 2, 2, 2/)
    array2 = 1.0_r_k
    array(:,2,2,2) = 1.0_r_k
    call fd4_util_put_array(d, 1, 1, abnd2, array2, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump
  
    abnd2(0:3,1) = (/ 1,    sz-ng(1), sz-ng(2), sz-ng(3)/)
    abnd2(0:3,2) = (/ bins, sz-ng(1), sz-ng(2), sz-ng(3)/)
    array(:, sz-ng(1), sz-ng(2), sz-ng(3)) = 1.0_r_k
    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

    call get_compare_dump

    deallocate(array2)

  else if(scenario==7) then

    array(:,sz/4:sz/4,sz/4:sz/2,sz/2:sz/2) = 1.0_r_k
    !array(:,sz/4:3*sz/4,sz/4:3*sz/4,sz/4:3*sz/4) = 1.0_r_k

    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length

!!! facevar stuff
#ifdef COMMENT
    array = 1.0_r_k
    call fd4_util_put_array(d, 2, 1, abnd, array, err)
    compare = 3.0_r_k
    call fd4_util_get_array(d, 2, 1, abnd, compare, err, 0)
    if(rank==0) then
      do k=1,aext(3)
        call dump_r(compare(1,:,:,:), 1,aext(1),1,aext(2),k,.true.)
      end do
    end if
#endif
!!! facevar stuff

    array = 0.0_r_k
    array(:,sz/4:sz/4,sz/4:sz/2,sz/2:sz/2) = 1.0_r_k
    call get_compare_dump
    
    array = 5.0_r_k
    call fd4_util_put_array(d, 3, 1, abnd, array, err)

    call fd4_domain_max_bext(d,wext(1:3),.true.)
    wext(0)=bins
    allocate(work(wext(0),wext(1),wext(2),wext(3)))
    work=0.0_r_k
    
    call mpi_barrier(comm,err)

    call fd4_ghostcomm_create(gc, d, 2, (/1,3/), (/1,1/), err, opt_exchdim=(/FD4_XYZ,FD4_XYZ/) )
    call fd4_ghostcomm_exch(gc, err)
    
    call fd4_iter_init(d, iter)
    do while(associated(iter%cur))
      work = -1.0_r_k
      call fd4_iter_get_ghost(iter,1,1,wext,work,bext)
      do k=1,bext(3)
        call dump_r(work(1,:,:,:),1,bext(1),1,bext(2),k,.true.)
      end do
      call fd4_iter_get_ghost(iter,3,1,wext,work,bext)
      do k=1,bext(3)
        call dump_r(work(1,:,:,:),1,bext(1),1,bext(2),k,.true.)
      end do
      call fd4_iter_next(iter)
    end do

    call fd4_ghostcomm_delete(gc)
    
  elseif(scenario==8 .and. nproc==2) then

    array(:,sz/4:sz/4,sz/4:sz/2,sz/2:sz/2) = 1.0_r_k

    call fd4_util_put_array(d, 1, 1, abnd, array, err)

    print *,rank,' number of blocks: ',d%blocks%length
    call get_compare_dump

    call fd4_couple_create(cpl, d, err, opt_nvar=2, opt_nabnd=2)
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)

#define CELLCENTERED
    do i=0,nproc-1
      cabnd(1:3,1) = (/ 1+(i*sz)/nproc,   1  , 1  /)
      cabnd(1:3,2) = (/ ((i+1)*sz)/nproc, sz , sz /)
#ifdef CELLCENTERED
      ! coupling to cell centered var
      call fd4_couple_add_array(cpl, i, 1, 1, cabnd, caidx(1), err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
#endif
      ! coupling to a face var
      !if(i==0) cabnd(1,2) = cabnd(1,2) - 1
      !if(i==1) cabnd(1,1) = cabnd(1,1) - 1
      call fd4_couple_add_array(cpl, i, 2, 1, cabnd, caidx(2), err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
    end do
    
#ifdef CELLCENTERED
    if(bins==1) then
      allocate(cplarray(1:sz/nproc,1:sz,1:sz))
      cplarray = REAL((rank+1)*3,r_k)
      cplarrayp => cplarray
      call fd4_couple_set_local_3D_array(cpl, caidx(1), cplarrayp, (/1,1,1/), err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
    else
      allocate(cplarray4(-4:2*bins,1:sz/nproc,1:sz,1:sz))
      cplarray4 = REAL((rank+1)*3,r_k)
      cplarray4p => cplarray4
      call fd4_couple_set_local_4D_array(cpl, caidx(1), cplarray4p, (/1,1,1/), -2, err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
    end if
#endif

    allocate(cplarrayf(1:sz+1,1:sz+1,1:sz))
    cplarrayf = REAL((rank+1)*10,r_k)
    cplarrayp => cplarrayf
    call fd4_couple_set_local_3D_array(cpl, caidx(2), cplarrayp, (/1,1,1/), err)
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)

    call fd4_couple_debug_metadata(cpl,'couple')
    
    call fd4_couple_commit(cpl, err)
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)

    if(rank==0) print *,'fd4_couple_put:'
    call fd4_couple_put(cpl, err)
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
#ifdef CELLCENTERED
    call get_dump(1)
#endif
    call get_dump(2)
    


#define COUPLE_FACE_STUFF
#ifdef COUPLE_FACE_STUFF

    call fd4_iter_init(d, iter)
    do while(associated(iter%cur))
      if(iter%cur%pos(2)==1 .and. iter%cur%pos(3)==1) then
        write(*,'(I6,A,3I4,A,2F5.0)') rank, ': ', iter%cur%pos, ': ', iter%cur%fields(2,1)%l(1,1:2,1,1)
      end if
      !write(*,'(I6,A,3I4,A,5F5.0)') rank, ': ', iter%cur%pos, ': ', iter%cur%fields(2,1)%l(1,1:5,1,1)
      call fd4_iter_next(iter)
    end do

#endif

    !call fd4_util_propagate_facevar (d, (/2/), (/1/), .true., err)
    !if(err/=0) call MPI_Abort(d%mpi%comm, err, err)

#define COUPLE_GET
#ifdef COUPLE_GET
    
    ! reverse coupling - fd4_couple_get
#ifdef CELLCENTERED
    if(bins==1) then
      cplarray = -1.0
    else
      cplarray4 = -1.0
    end if
#endif
    cplarrayf = -1.0
    if(rank==0) print *,'fd4_couple_get:'
    call fd4_couple_get(cpl, err, opt_novnull = .false.)    
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
#ifdef CELLCENTERED
    do k=1,sz
      if(bins==1) then
        call dump_r(cplarray, 1,sz/nproc,1,sz, k,.true.)
      else
        call dump_r(cplarray4(1,:,:,:),  1,sz/nproc,1,sz, k,.true.)
      end if
    end do
#endif
    do k=1,sz
      call dump_r(cplarrayf,  1, sz+1, 1, sz+1, k,.true.)
    end do
!COUPLE_GET
#endif

    call fd4_couple_delete(cpl, err)
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
    
    !call fd4_util_check_facevar(d, 2, 1, err)
    
  elseif(scenario==9 .and. nproc==2) then
    ! 2009-07-08
    ! coupling with mark_blocks - balance_readjust - couple_put

    array(:,sz/4:sz/4,sz/4:sz/2,sz/2:sz/2) = 1.0_r_k
    call fd4_util_put_array(d, 1, 1, abnd, array, err)
    print *,rank,' number of blocks: ',d%blocks%length
    call get_compare_dump

    call fd4_couple_create(cpl, d, err)  ! opt_nvar, opt_nabnd
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)

    do i=0,nproc-1
      ! coupling to cell centered var
      cabnd(1:3,1) = (/ 1+(i*sz)/nproc,   1  , 1  /)
      cabnd(1:3,2) = (/ ((i+1)*sz)/nproc, sz , sz /)
      call fd4_couple_add_array(cpl, i, 1, 1, cabnd, caidx(1), err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
      ! coupling to a face var
      cabnd(1:3,1) = (/ 1,    1+(i*sz)/nproc,   1  /)
      cabnd(1:3,2) = (/ sz,   ((i+1)*sz)/nproc, sz /)      
      call fd4_couple_add_array(cpl, i, 2, 1, cabnd, caidx(2), err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
    end do

    if(bins==1) then
      allocate(cplarray(1:sz/nproc,1:sz,1:sz))
      cplarray = REAL((rank+1)*3,r_k)
      cplarrayp => cplarray
      call fd4_couple_set_local_3D_array(cpl, caidx(1), cplarrayp, (/1,1,1/), err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
    else
      allocate(cplarray4(-4:2*bins,1:sz/nproc,1:sz,1:sz))
      cplarray4 = REAL((rank+1)*3,r_k)
      cplarray4p => cplarray4
      call fd4_couple_set_local_4D_array(cpl, caidx(1), cplarray4p, (/1,1,1/), -2, err)
      if(err/=0) call MPI_Abort(d%mpi%comm, err, err)
    end if

    allocate(cplarrayf(1:sz+1,1:sz+1,1:sz))
    cplarrayf = REAL((rank+1)*10,r_k)
    cplarrayp => cplarrayf
    call fd4_couple_set_local_3D_array(cpl, caidx(2), cplarrayp, (/1,1,1/), err)
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)

    call fd4_couple_debug_metadata(cpl, 'couple')
    
    call fd4_couple_commit(cpl, err)
    if(err/=0) call MPI_Abort(d%mpi%comm, err, err)

    if(rank==0) print *,'fd4_couple_mark_blocks:'
    call fd4_couple_mark_blocks(cpl, err)    
    print *,rank,' number of blocks: ',d%blocks%length
    print *,rank,' number of marked blocks: ',d%markedblocks%pos
    !call get_dump(1)
    !call get_dump(2)
    
    call fd4_balance_readjust(d, err)
    if(rank==0) print *,'fd4_balance_readjust:'
    print *,rank,' number of blocks: ',d%blocks%length

    if(rank==0) print *,'fd4_couple_put:'
    call fd4_couple_put(cpl, err)
    call fd4_couple_delete(cpl, err)
    call get_dump(1)
    call get_dump(2)
    

  end if ! scenario

  call MPI_Barrier(comm,err)


  str=''
  call fd4_iter_init(d, iter)
  do while(associated(iter%cur))
    write(str2,'(I10)') iter%biter%cur%key  
    str = STRIM(str)//STRIM(str2)
    call fd4_iter_next(iter)
  end do  

  if(rank==0) print *,rank,': ',str



#ifdef FD4_VIS5D
  call fd4_vis5d_close(err)
#endif  




  deallocate(array)


  call fd4_domain_delete(d)

contains
    
  subroutine get_compare_dump
    integer(i_k) :: k

    compare = 0.0_r_k
    call fd4_util_get_array(d, 1, 1, abnd, compare, err, 0)

    if(rank==0 .and. dump) then
      do k=abnd(3,1),abnd(3,2)
        call dump_r(compare(1,:,:,:), abnd(1,1),abnd(1,2),abnd(2,1),abnd(2,2),k,.true.)
      end do
      do k=2,bsize(3)+1
        call dump_i(d%blkmap(:,:,:,d%blkmap_now), 2, bsize(1)+1, 2, bsize(2)+1, k)
      end do
    end if
    
    compare = array - compare
    if( rank == 0 .and. (MAXVAL(compare)/=0.0 .or. MINVAL(compare)/=0.0) ) then
      print *,'ERROR! Difference is not 0:'
      do k=abnd(3,1),abnd(3,2)
        call dump_r(compare(1,:,:,:), abnd(1,1),abnd(1,2),abnd(2,1),abnd(2,2),k,.true.)
      end do
    end if

#ifdef FD4_VIS5D
    call fd4_vis5d_write(err)
#endif  

  end subroutine get_compare_dump


  subroutine get_dump(var)
    integer(i_k), intent(in) :: var
    integer(i_k) :: k
    compare = 0.0_r_k
    call fd4_util_get_array(d, var, 1, abnd, compare, err, 0)
    if(rank==0 .and. dump) then
      do k=abnd(3,1),abnd(3,2)
        call dump_r(compare(1,:,:,:), abnd(1,1),abnd(1,2),abnd(2,1),abnd(2,2),k,.true.)
      end do
    end if
  end subroutine get_dump

  
  subroutine mkcloud(s,v,x,y,z,r)

    REAL (KIND=r_k) , INTENT(INOUT)  ::  s(:,:,:,:)
    REAL (KIND=r_k) , INTENT(IN)  ::  v
    REAL,  intent(in) :: r,x,y,z
    integer(i_k) :: i,j,k
  
    ! initialize s with a spheric thing
    do k=1,ubound(s,4)
      do j=1,ubound(s,3)
        do i=1,ubound(s,2)
          if(sqrt(abs(i-x)**2+abs(j-y)**2+abs(k-z)**2)<r) s(:,i,j,k) = v
        end do
      end do
    end do
  
  end subroutine mkcloud
  
  

  subroutine dump_r(s,istart, iend, jstart, jend, k, number)

    INTEGER,  INTENT (IN) ::    k, istart, iend, jstart, jend    
    REAL    (KIND=r_k) , INTENT(IN)  ::  s(:,:,:)  
    logical, optional, intent(in) :: number
    integer :: i,j
    logical :: num
    
    num=.false.
    if(present(number)) num=number

    print *,'------'

    do j=jstart,jend
      do i=istart,iend
        if(num) then
          write(*,'(F4.0)',advance='no') s(i,j,k)
          !write(*,'(E9.2)',advance='no') s(i,j,k)
          !write(*,'(E15.8)',advance='no') s(i,j,k)
        else
          !
          !  M == s >  0.99
          !  m == s >  0.1
          !  o == s >  1e-20
          !  . == s >  0
          !    == s =  0
          !  - == s <  0
          !
          if(s(i,j,k)>0.99) then
            write(*,'(A2)',advance='no') 'M'
          elseif(s(i,j,k)>0.1) then
            write(*,'(A2)',advance='no') 'm'
          elseif(s(i,j,k)>1e-20) then
            write(*,'(A2)',advance='no') 'o'
          elseif(s(i,j,k)>0) then
            write(*,'(A2)',advance='no') '.'
          elseif(s(i,j,k)==0) then
            write(*,'(A2)',advance='no') ' '
          else
            write(*,'(A2)',advance='no') '-'
          end if
        end if
      end do
      write(*,*) 
    end do

    !print *,'------'

  end subroutine


  subroutine dump_i(s,istart, iend, jstart, jend, k)

    INTEGER,  INTENT (IN) ::    k, istart, iend, jstart, jend    
    integer  (KIND=i_k) , INTENT(IN)  ::  s(:,:,:)  
    integer :: i,j
    
    print *,'------'

    do j=jstart,jend
      do i=istart,iend
        if(s(i,j,k)<0) then
          write(*,'(A4)',advance='no') '   .'
        else
          write(*,'(I4)',advance='no') s(i,j,k)
        end if
      end do
      write(*,*) 
    end do

    !print *,'------'

  end subroutine


end subroutine test_put_array
  
  
